@@ -26,7 +26,7 @@ import errno
 import os.path
 import subprocess
 import tempfile
-from typing import (Dict, FrozenSet, Iterable, List, Optional, Union)
+from typing import (Dict, FrozenSet, Iterable, List, Optional, Tuple, Union)
 
 import debian.deb822
 import debian.debian_support
@@ -51,6 +51,7 @@ known_restrictions = frozenset({
     'needs-root',
     'needs-sudo',
     'rw-build-tree',
+    'skip-foreign-architecture',
     'skip-not-installable',
     'skippable',
     'superficial',
@@ -81,6 +82,9 @@ RESTRICTIONS_IMPLY_DEPS = {
 }
 
 
+PKGDATADIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
+
+
 class Unsupported(Exception):
     '''Test cannot be run in the testbed'''
 
@@ -123,7 +127,7 @@ class Test:
         restrictions: Iterable[str],
         features: Iterable[str],
         depends: List[str],
-        synth_depends: List[str],
+        synth_depends: List[List[str]],
     ) -> None:
         '''Create new test description
 
@@ -191,12 +195,18 @@ class Test:
         else:
             adtlog.report(self.name, 'FAIL ' + reason)
 
-    def check_testbed_compat(self, caps, ignore_restrictions=()):
+    def check_testbed_compat(
+        self,
+        caps: List[str],
+        ignore_restrictions: Optional[List[str]] = None,
+    ) -> None:
         '''Check for restrictions incompatible with test bed capabilities.
 
         Raise Unsupported exception if there are any.
         '''
-        effective = set(self.restrictions) - set(ignore_restrictions)
+        effective = set(self.restrictions)
+        if ignore_restrictions:
+            effective -= set(ignore_restrictions)
         provided = set(caps)
 
         for r in effective:
@@ -283,9 +293,45 @@ def _debian_check_unknown_fields(name, record):
         raise Unsupported(name, 'unknown field %s' % unknown_keys.pop())
 
 
-def _debian_packages_from_source(srcdir, foreign_test_arch=None):
-    packages = []
-    packages_no_arch = []
+def _architecture_is_concerned(
+    architecture: str,
+    architecture_restrictions: List[str],
+) -> bool:
+    """Wrapper around python-debian's architecture_is_concerned() with perl
+    fallback in case python-debian is too old."""
+
+    # TODO: Add Depends: python3-debian (>= 0.1.48) to bin:autopkgtest and drop this.
+
+    if hasattr(debian.debian_support, "DpkgArchTable") and hasattr(
+        debian.debian_support.DpkgArchTable, "architecture_is_concerned"
+    ):
+        dpkg_arch_table = debian.debian_support.DpkgArchTable.load_arch_table()
+        return dpkg_arch_table.architecture_is_concerned(
+            architecture,
+            architecture_restrictions,
+        )
+    else:
+        res = subprocess.run(
+            [
+                os.path.join(PKGDATADIR, "lib", "arch-is-concerned.pl"),
+                architecture,
+                " ".join(architecture_restrictions),
+            ]
+        )
+        return res.returncode == 0
+
+
+def _debian_packages_from_source(
+    srcdir: str,
+    test_arch: str,
+    test_arch_is_foreign: bool,
+) -> Tuple[List[str], List[str]]:
+    """Parse source tree in srcdir and return tuple with: list of all binary
+    deb packages built from that source, and the same list but restricted to
+    packages that build for test_arch."""
+
+    packages: List[str] = []
+    packages_for_test_arch: List[str] = []
 
     for st in parse_rfc822(os.path.join(srcdir, 'debian/control')):
         if 'Package' not in st:
@@ -295,59 +341,42 @@ def _debian_packages_from_source(srcdir, foreign_test_arch=None):
         if st.get('Xc-package-type', 'deb') != 'deb' or \
                 st.get('Package-type', 'deb') != 'deb':
             continue
+        binarydeb = st['Package']
         arch = st['Architecture']
-        qual_pkg = st['Package']
-        # take care to emit an arch qualifier only for arch-dependent
-        # packages, not for arch: all ones
-        if foreign_test_arch:
-            qual_pkg += ':' + foreign_test_arch
+        packages.append(binarydeb)
         if arch == 'all':
-            packages.append(st['Package'])
-        elif arch == 'any':
-            packages.append(qual_pkg)
+            packages_for_test_arch.append(binarydeb)
         else:
-            packages.append('%s [%s]' % (qual_pkg, arch))
-        packages_no_arch.append(st['Package'])
+            arch_list = arch.split()
+            if _architecture_is_concerned(test_arch, arch_list):
+                if test_arch_is_foreign:
+                    binarydeb += ':' + test_arch
+                packages_for_test_arch.append(binarydeb)
 
-    return (packages, packages_no_arch)
+    return (packages, packages_for_test_arch)
 
 
-def _filter_variables(deps):
+def _filter_variables(deps: Iterable[str]) -> List[str]:
     deplist = []
-    for dep in [d.strip() for d in deps.split(',')]:
+    for dep in deps:
         if re.match(r'^\${', dep):
             continue
         dep = re.sub(r'\(.*\${\S*}.*\)', '', dep)
-        deplist.append(dep)
-    return ', '.join(deplist)
+        deplist.append(dep.strip())
+    return deplist
 
 
-def _debian_recommends_from_source(srcdir, testbed_arch):
-    deps = ''
+def _debian_recommends_from_source(srcdir: str) -> List[str]:
+    deps = []
     for st in parse_rfc822(os.path.join(srcdir, 'debian/control')):
         if 'Recommends' in st:
-            deps += st['Recommends'] + ','
+            deps += [d.strip() for d in st['Recommends'].strip(',').split(',')]
 
     # Because we parse the source d/control file instead of the binary
     # control file, we may run into variables. See bug 1008206 for
     # some discussion.
     deps = _filter_variables(deps)
 
-    # resolve arch specific dependencies and build profiles
-    perl = subprocess.Popen(['perl', '-'], stdin=subprocess.PIPE,
-                            stdout=subprocess.PIPE)
-    code = '''use Dpkg::Deps;
-              $dep = deps_parse('%s', reduce_restrictions => 1,
-                                host_arch => '%s');
-              $out = $dep->output();
-              print $out, "\\n";
-              ''' % (deps, testbed_arch)
-    deps = perl.communicate(code.encode('UTF-8'))[0].decode('UTF-8').strip()
-    if perl.returncode != 0:
-        raise InvalidControl('source', 'Invalid Recommends')
-
-    deps = [d.strip() for d in deps.split(',')]
-
     # The following check is a bit silly as there were apparently no
     # recommends, but still the test asked for it. Let's not fail on
     # that, but also let's not add the empty string.
@@ -357,44 +386,15 @@ def _debian_recommends_from_source(srcdir, testbed_arch):
     return deps
 
 
-def _debian_build_deps_from_source(srcdir, testbed_arch):
-    deps = ''
+def _debian_build_deps_from_source(srcdir: str) -> List[str]:
+    deps = []
     for st in parse_rfc822(os.path.join(srcdir, 'debian/control')):
-        if 'Build-depends' in st:
-            deps += st['Build-depends']
-        if 'Build-depends-indep' in st:
-            deps += ', ' + st['Build-depends-indep']
-        if 'Build-depends-arch' in st:
-            deps += ', ' + st['Build-depends-arch']
-
-    # resolve arch specific dependencies and build profiles
-    perl = subprocess.Popen(['perl', '-'], stdin=subprocess.PIPE,
-                            stdout=subprocess.PIPE)
-    code = '''use Dpkg::Deps;
-              $supports_profiles = ($Dpkg::Deps::VERSION gt '1.04');
-              $dep = deps_parse('%s', reduce_arch => 1,
-                                reduce_profiles => $supports_profiles,
-                                build_dep => 1, host_arch => '%s');
-              $out = $dep->output();
-              # fall back to ignoring build profiles
-              $out =~ s/ <[^ >]+>//g if (!$supports_profiles);
-              print $out, "\\n";
-              ''' % (deps, testbed_arch)
-    deps = perl.communicate(code.encode('UTF-8'))[0].decode('UTF-8').strip()
-    if perl.returncode != 0:
-        raise InvalidControl('source', 'Invalid build dependencies')
-
-    deps = [d.strip() for d in deps.split(',')]
+        for key in ('Build-depends', 'Build-depends-indep', 'Build-depends-arch'):
+            if key in st:
+                deps += [d.strip() for d in st[key].strip(',').split(',')]
 
     # @builddeps@ should always imply build-essential
-    #
-    # Ensure we are installing the testbed-native version of build-essential
-    # by specifying :native. This is needed because the build-essential package
-    # is not Multi-Arch: foreign, so we need to make sure we are always picking
-    # up the package for the testbed architecture, not for a cross-test
-    # architecture.
-    deps.append('build-essential:native')
-
+    deps.append('build-essential')
     return deps
 
 
@@ -426,54 +426,30 @@ def _debian_check_dep(testname, dep):
     return (m.group('package'), m.group('version'))
 
 
-def _synthesize_deps(dep, test_arch):
-    '''Ensure that apt can install synthesized Depends
+def _expand_test_depends(
+    testname: str,
+    dep_str: str,
+    srcdir: str,
+    test_arch: str,
+    test_arch_is_foreign: bool,
+) -> Tuple[List[str], List[List[str]]]:
+    '''Split test dependencies (comma separated), validate their syntax, and
+    expand @, @builddeps@ and @recommends@. Returns list of dependencies
+    and list of synthesised dependencies.
 
-    Test Depends may have architecture qualifiers we need to check,
-    because apt command-line can't handle those and can't do the check.
-    (Policy says the architecture qualifier should not be in the binary
-    control file). We'll ignore the version here.
-    '''
-
-    dep = dep.strip()
-    m = dep_re.match(dep)
-
-    arch_matches = 'y'
-    if m.group('arch') is not None:
-        arch_matches = False
-        perl = subprocess.Popen(['perl', '-'], stdin=subprocess.PIPE,
-                                stdout=subprocess.PIPE)
-        code = '''use Dpkg::Deps;
-                  print deps_parse('foo [%s]')->arch_is_concerned('%s') ? 'y' : 'n';
-                  ''' % (m.group('arch'), test_arch)
-        arch_matches = perl.communicate(code.encode('UTF-8'))[0].decode('UTF-8').strip()
-        if perl.returncode != 0:
-            raise InvalidControl('source', 'Invalid (Test-)Depends architecture qualifiers')
-
-    if arch_matches == 'y':
-        return m.group('package')
-    else:
-        return None
-
-
-def _parse_debian_depends(testname, dep_str, srcdir, testbed_arch,
-                          foreign_test_arch=None):
-    '''Parse Depends: line in a Debian package
-
-    Split dependencies (comma separated), validate their syntax, and expand @,
-    @builddeps@ and @recommends@. Return a list of dependencies.
+    If test_arch_is_foreign is True, then :test_arch architecture qualifiers
+    are added to packages that are not Architecture: all.
 
     This may raise an InvalidControl exception if there are invalid
     dependencies.
     '''
     deps = []
     synthdeps = []
-    (my_packages, my_packages_no_arch) = _debian_packages_from_source(srcdir,
-                                                                      foreign_test_arch)
-    if foreign_test_arch:
-        test_arch = foreign_test_arch
-    else:
-        test_arch = testbed_arch
+    (my_packages, my_packages_for_test_arch) = _debian_packages_from_source(
+        srcdir,
+        test_arch,
+        test_arch_is_foreign,
+    )
 
     for alt_group_str in dep_str.split(','):
         alt_group_str = alt_group_str.strip()
@@ -482,60 +458,47 @@ def _parse_debian_depends(testname, dep_str, srcdir, testbed_arch,
             continue
         adtlog.debug('processing dependency %s' % alt_group_str)
         if alt_group_str == '@':
-            for d in my_packages:
+            for d in my_packages_for_test_arch:
                 adtlog.debug('synthesised dependency %s' % d)
                 deps.append(d)
-                s = _synthesize_deps(d, test_arch)
-                if s:
-                    synthdeps.append(s)
+                synthdeps.append([d])
         elif alt_group_str == '@builddeps@':
-            for d in _debian_build_deps_from_source(srcdir, testbed_arch):
+            for d in _debian_build_deps_from_source(srcdir):
                 adtlog.debug('synthesised dependency %s' % d)
                 deps.append(d)
         elif alt_group_str == '@recommends@':
-            for d in _debian_recommends_from_source(srcdir, testbed_arch):
+            for d in _debian_recommends_from_source(srcdir):
                 adtlog.debug('synthesised dependency %s' % d)
                 deps.append(d)
-        elif alt_group_str == 'build-essential':
-            # This is how packages declare they want to build code during the
-            # test (but not necessarily doing a full package build), but for
-            # foreign-architecture testing this declaration is wrong because
-            # it tries to pull in build-essential for the wrong arch. We need
-            # to explicitly request the :native version.
-            deps.append('build-essential:native')
-            # Install packages needed for cross-building.
-            if foreign_test_arch:
-                deps.append('crossbuild-essential-%s:native' % foreign_test_arch)
-                # also add build-dependencies on libc-dev and libstdc++-dev,
-                # as a workaround to #815172.
-                deps.append('libc-dev:%s' % foreign_test_arch)
-                deps.append('libstdc++-dev:%s' % foreign_test_arch)
         else:
+            alt_group_list = []
             synthdep_alternatives = []
-            for dep in alt_group_str.split('|'):
+            for dep in [d.strip() for d in alt_group_str.split('|')]:
                 (pkg, version) = _debian_check_dep(testname, dep)
-                if pkg not in my_packages_no_arch:
-                    synthdep_alternatives = []
-                    break
-                s = _synthesize_deps(dep, test_arch)
-                if s:
-                    if foreign_test_arch:
-                        s += ':' + foreign_test_arch
-                    synthdep_alternatives.append(s)
-            if synthdep_alternatives:
+                if test_arch_is_foreign and f'{pkg}:{test_arch}' in my_packages_for_test_arch:
+                    dep = dep.replace(pkg, f'{pkg}:{test_arch}', 1)
+                    pkg += f':{test_arch}'
+                alt_group_list.append(dep)
+                if pkg in my_packages:
+                    synthdep_alternatives.append(pkg)
+            if len(alt_group_list) == len(synthdep_alternatives):
                 adtlog.debug('marked alternatives %s as a synthesised dependency' % synthdep_alternatives)
-                if len(synthdep_alternatives) > 1:
-                    synthdeps.append(synthdep_alternatives)
-                else:
-                    synthdeps.append(synthdep_alternatives[0])
-            if foreign_test_arch:
-                for mine in my_packages:
-                    # ignore [arch] filters for matching packages
-                    if alt_group_str + ':' + foreign_test_arch == mine.split(' ')[0]:
-                        adtlog.debug('%s is from our source package, adding arch qualifier for cross-testing' % s)
-                        alt_group_str += ':' + foreign_test_arch
-                        break
-            deps.append(alt_group_str)
+                synthdeps.append(synthdep_alternatives)
+            deps.append(' | '.join(alt_group_list))
+
+    if test_arch_is_foreign and "build-essential" in deps:
+        # Ensure we are installing the testbed-native version of
+        # build-essential by specifying :native. This is needed because the
+        # build-essential package is not Multi-Arch: foreign, so we need to
+        # make sure we are always picking up the package for the testbed
+        # architecture, not for the foreign test architecture.
+        deps.remove('build-essential')
+        deps.append('build-essential:native')
+        # Install the right crossbuild-essential package.
+        deps.append('crossbuild-essential-%s:native' % test_arch)
+        # Add install libc-dev and libstdc++-dev, as a workaround for #815172.
+        deps.append('libc-dev:%s' % test_arch)
+        deps.append('libstdc++-dev:%s' % test_arch)
 
     return (deps, synthdeps)
 
@@ -579,8 +542,8 @@ def _matches_architecture(host_arch, arch_wildcard):
     return True
 
 
-def _check_architecture(name, testbed_arch, architectures):
-    '''Check if testbed_arch is supported by the architectures
+def _check_architecture(name, test_arch, architectures):
+    '''Check if test_arch is supported by the test architectures
 
     The architecture list comes in two variants, positive: only this
     arch is supported (arch may be a wildcard) and negative: this arch
@@ -603,9 +566,9 @@ def _check_architecture(name, testbed_arch, architectures):
     for arch in architectures:
         if arch[0] == "!":
             any_negative = True
-            if _matches_architecture(testbed_arch, arch[1:]):
+            if _matches_architecture(test_arch, arch[1:]):
                 raise Unsupported(name, "Test declares architecture as not " +
-                                  "supported: %s" % testbed_arch)
+                                  "supported: %s" % test_arch)
         if arch[0] != "!":
             any_positive = True
 
@@ -616,18 +579,25 @@ def _check_architecture(name, testbed_arch, architectures):
                               "others aren't")
         arch_matched = False
         for arch in architectures:
-            if _matches_architecture(testbed_arch, arch):
+            if _matches_architecture(test_arch, arch):
                 arch_matched = True
 
         if not arch_matched:
             raise Unsupported(name, "Test lists explicitly supported " +
                               "architectures, but the current architecture " +
-                              "%s isn't listed." % testbed_arch)
+                              "%s isn't listed." % test_arch)
 
 
-def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
-                        auto_control=True, foreign_test_arch=None,
-                        ignore_restrictions=(), only_tests=()):
+def parse_debian_source(
+    srcdir: str,
+    testbed_caps: List[str],
+    test_arch: str,
+    test_arch_is_foreign: bool,
+    control_path: Optional[str] = None,
+    auto_control: bool = True,
+    ignore_restrictions: Optional[List[str]] = None,
+    only_tests: Optional[List[str]] = None,
+) -> Tuple[List[Test], bool]:
     '''Parse test descriptions from a Debian DEP-8 source dir
 
     @ignore_restrictions: If we would skip the test due to these restrictions,
@@ -646,6 +616,8 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
     some_skipped = False
     command_counter = 0
     tests = []
+    if not ignore_restrictions:
+        ignore_restrictions = []
     if not control_path:
         control_path = os.path.join(srcdir, 'debian', 'tests', 'control')
         dtc_exists = os.path.exists(control_path)
@@ -710,12 +682,13 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
                 test_names = record['Tests'].replace(',', ' ').split()
                 if len(test_names) == 0:
                     raise InvalidControl('*', '"Tests" field is empty')
-                (depends, synth_depends) = _parse_debian_depends(
+                (depends, synth_depends) = _expand_test_depends(
                     test_names[0],
                     record.get('Depends', '@'),
                     srcdir,
-                    testbed_arch,
-                    foreign_test_arch)
+                    test_arch,
+                    test_arch_is_foreign,
+                )
                 if 'Test-command' in record:
                     raise InvalidControl('*', 'Only one of "Tests" or '
                                          '"Test-Command" may be given')
@@ -724,11 +697,6 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
                         '*', 'test-name feature incompatible with Tests')
                 test_dir = record.get('Tests-directory', 'debian/tests')
 
-                if foreign_test_arch:
-                    test_arch = foreign_test_arch
-                else:
-                    test_arch = testbed_arch
-
                 for n in test_names:
                     try:
                         _debian_check_unknown_fields(n, record)
@@ -745,19 +713,20 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
                         tests.append(test)
             elif 'Test-command' in record:
                 command = record['Test-command']
-                (depends, synth_depends) = _parse_debian_depends(
+                (depends, synth_depends) = _expand_test_depends(
                     command,
                     record.get('Depends', '@'),
                     srcdir,
-                    testbed_arch,
-                    foreign_test_arch)
+                    test_arch,
+                    test_arch_is_foreign,
+                )
                 if feature_test_name is None:
                     command_counter += 1
                     name = 'command%i' % command_counter
                 else:
                     name = feature_test_name
                 _debian_check_unknown_fields(name, record)
-                _check_architecture(name, testbed_arch, architectures)
+                _check_architecture(name, test_arch, architectures)
                 test = Test(name, None, command, restrictions, features,
                             depends, synth_depends)
                 test.check_testbed_compat(testbed_caps, ignore_restrictions)
